/**********************************************************************/
/* CellSim/LSView v0.3.0              2008.03.28 Arch Lab. TOKYO TECH */
/**********************************************************************/

import javax.swing.*;
import javax.swing.event.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;

/**********************************************************************/
public class lsview extends JFrame implements ActionListener, ChangeListener{
    static final long serialVersionUID = 1;
    JSlider sliValue;
    JLabel lblValue;
    BMPPanel pnlBMP[];
    JButton cmdAuto;
    LoadData ld;
    AutoInc inc;
    static String filename;
    static LogReader lr;
    static LogArray log[][];
    static int maxspe;
    static int maxpage;
    static int col_mul;

    lsview(){
        System.out.println("preparing window...");

        // Frame constructor: initialize objects
        sliValue = new JSlider(0, maxpage - 1, 0);
        sliValue.addChangeListener(this);
        lblValue = new JLabel("0");
        lblValue.setPreferredSize(new Dimension(50, 20));
        pnlBMP = new BMPPanel[maxspe];
        for (int i = 0; i < maxspe; i++) {
            pnlBMP[i] = new BMPPanel();
            pnlBMP[i].Panel.setPreferredSize(new Dimension(258, 258));
            pnlBMP[i].setImage(log[i][0]);
        }
        cmdAuto = new JButton("Start");
        cmdAuto.addActionListener(this);

        // add objects to panel
        JPanel base = new JPanel();
        JPanel p = new JPanel();
        p.setLayout(new FlowLayout());
        p.setPreferredSize(new Dimension(550, 65 + (maxspe + 1) / 2 * 270));
        for (int i = 0; i < pnlBMP.length; i++)
            p.add(pnlBMP[i].Panel);
        base.add(sliValue);
        base.add(lblValue);
        base.add(cmdAuto);
        base.add(p);
        getContentPane().add(base, BorderLayout.CENTER);

        this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        this.setTitle("SimCell/LSView");
        this.setBounds(10, 10, 600, 85 + (maxspe + 1) / 2 * 270);

        ld = new LoadData();
        ld.start();
    }

    public static void main(String[] args) {
        if (!checkArg(args))
            return;
        if (!(checkHeader()))
            return;
        log = new LogArray[maxspe][maxpage];
        for (int i = 0; i < maxspe; i++)
            for (int j = 0; j < maxpage; j++)
                log[i][j] = new LogArray(0);

        JFrame frame = new lsview();
        frame.setVisible(true);
    }       

    static boolean checkArg(String[] args) {
        // check arguments
        if ((args.length <= 0) || (args.length >= 3)) {
            usage();
            return false;
        }
        if (args.length == 2) {
            if (args[0].substring(0, 2).equals("-c")) {
                try {
                    col_mul = Integer.parseInt(args[0].substring(2));
                } catch (NumberFormatException e) {}
                if (col_mul < 1) {
                    System.err.println("** invalid expression of -c");
                    usage();
                    return false;
                }
            } else {
                System.err.println("** invalid option");
                usage();
                return false;
            }
        } else {
            col_mul = 4;
        }
        filename = args[args.length - 1];
        if (!(new File(filename).isFile())) {
            System.err.println(filename + " isn't file");
            return false;
        }
        return true;
    }

    static void usage() {
        System.err.println("usage: lsview [-cnum] filename");
        System.err.println("-cnum -- color depth coefficient");
    }

    static boolean checkHeader() {
        System.out.println("checking data...");
        lr = new LogReader(filename, col_mul);
        if ((!lr.available) || (lr.maxspe == 0)) {
            System.err.println("** invalid data");
            return false;
        }
        maxspe = lr.maxspe;
        maxpage = lr.maxpage;
        System.out.println("* Total SPEs: " + maxspe);
        System.out.println("* Total Pages: " + maxpage);
        return true;
    }

    public void actionPerformed(ActionEvent e) {
        // on cmdAuto_Click .. Start/Stop Thread AutoInc
        if (cmdAuto.getText().equals("Start")) {
            if (sliValue.getValue() == sliValue.getMaximum())
                sliValue.setValue(0);
            cmdAuto.setText("Stop");
            inc = new AutoInc();
            inc.start();
        } else {
            inc.StopFlag = true;
        }
    }

    public void stateChanged(ChangeEvent e) {
        // on sliValue_Change
        lblValue.setText(String.valueOf(sliValue.getValue()));
        for (int i = 0; i < maxspe; i++)
            pnlBMP[i].setImage(log[i][sliValue.getValue()]);
        repaint();
    }

    // Thread LoadData: load data on background
    class LoadData extends Thread {
        public void run() {
            System.out.println("loading data...");
            for (int j = 0; j < maxpage; j++) {
                for (int i = 0; i < maxspe; i++)
                    log[i][j] = lr.readPage();
                if (j == sliValue.getValue()) {
                    for (int i = 0; i < maxspe; i++)
                        pnlBMP[i].setImage(log[i][j]);
                    repaint();
                }
            }
            lr.close();
            System.out.println("load data complete");
        }
    }
        
    // Thread AutoInc: increment value of slider every 100msec
    class AutoInc extends Thread {
        boolean StopFlag;

        public AutoInc() {}
        public void run() {
            int v;
            
            StopFlag = false;
            while(true) {
                v = sliValue.getValue();
                if ((v == sliValue.getMaximum()) || (StopFlag))
                    break;
                v++;
                sliValue.setValue(v);

                try {
                    Thread.sleep(100);
                } catch (Exception e) {}
            }
            cmdAuto.setText("Start");
        }
    }
}

/**********************************************************************/
class BMPPanel {
    BMPPanelActual Panel;
    Image CurrentImg;

    BMPPanel() {
        Panel = new BMPPanelActual();
    }

    public void setImage(LogArray log) {
        Panel.setImage(log);
    }

    class BMPPanelActual extends JPanel {
        static final long serialVersionUID = 1;
        Image imgBuffer;
        Graphics gBuffer;
        LogArray log;

        public void setImage(LogArray src) {
            if (src != null)
                log = src;
            if (gBuffer == null)
                return;
            gBuffer.setColor(Color.BLACK);
            gBuffer.drawRect(0, 0, 257, 257);
            if (log.available) {
                for (int x = 0; x < 128; x++) {
                    for (int y = 0; y < 128; y++) {
                        gBuffer.setColor(new Color(log.getColor(x, y)));
                        gBuffer.fillRect(x * 2 + 1, y * 2 + 1, 2, 2);
                    }
                }
            } else {
                gBuffer.setColor(Color.BLACK);
                gBuffer.drawString("Loading...", 50, 50);
            }
        }
        public void update(Graphics g) {
            paint(g);
        }
        public void paint(Graphics g) {
            if (imgBuffer == null) {
                imgBuffer = createImage(258, 258);
                gBuffer = imgBuffer.getGraphics();
                setImage(null);
            }
            g.drawImage(imgBuffer, 0, 0, this);
        }
    } 
}

/**********************************************************************/
class LogReader {
    private FileInputStream fs;
    private int col_mul;
    boolean available;
    int maxpage;
    int maxspe;

    LogReader(String filename, int col_mul) {
        this.col_mul = col_mul;
        available = false;
        try {
            fs = new FileInputStream(filename);
            available = checkValid();
            if (available) {
                maxspe = read4Byte();
                maxpage = read4Byte();
                fs.skip(24);
            }
        } catch (IOException e) {}
    }

    public void close() {
        try {
            fs.close();
        } catch (IOException e) {}
    }

    private boolean checkValid() {
        byte[] b = new byte[32];
        final String correct = "SimCell_LS_LOG_FILE_VERSION_1.0\n";
        try {
            fs.read(b);
            return (new String(b, "US-ASCII").equals(correct));  
        } catch (IOException e) {}

        return false;
    }

    private int read4Byte() throws IOException {
        int big, little;
        byte[] b = new byte[4];
        fs.read(b);
        big =    (((b[3] & 0xff) << 24) | ((b[2] & 0xff) << 16) |
                  ((b[1] & 0xff) << 8) | (b[0] & 0xff)) & ~0x80000000;
        little = (((b[0] & 0xff) << 24) | ((b[1] & 0xff) << 16) |
                  ((b[2] & 0xff) << 8) | (b[3] & 0xff)) & ~0x80000000;
        return (big < little) ? big : little;
    }

    public LogArray readPage() {
        byte table[] = new byte[256];
        LogArray page = new LogArray(col_mul);

        page.initialize();
        try {
            fs.read(table);
            for (int i = 0; i < 4; i++)
                for (int j = 0; j < 64; j++)
                    if (table[i * 64 + j] != 0) {
                        fs.read(page.cnt[i][j * 2]);
                        fs.read(page.cnt[i][j * 2 + 1]);
                    }
            page.setColorArray();
        } catch (IOException e) {}
        return page;
    }
}

/**********************************************************************/
class LogArray {
    byte cnt[][][];
    int color[][];
    boolean available;
    private int col_mul;
    final static int READ = 0;
    final static int WRITE = 1;
    final static int DMA = 2;
    final static int INST = 3;
    
    LogArray(int col_mul) {
        this.col_mul = col_mul;
        available = false;
    }
    public void initialize() {
        cnt = new byte[4][128][128];
    }

    public void setColorArray() {
        color = new int[128][128];
        for (int x = 0; x < 128; x++) {
            for (int y = 0; y < 128; y++) {
                if (cnt[INST][y][x] != 0) {
                    int tmp = 383 - (((int) cnt[INST][y][x]) & 0xff) * col_mul;
                    tmp = (tmp < 0) ? 0 : tmp;
                    color[x][y] = (tmp < 128) ?
                        (tmp + 128) * 0x010100 :
                        (tmp - 128) + 0xffff00;
                } else {
                    int blue = (((int) cnt[READ][y][x]) & 0xff) * col_mul;
                    int red = (((int) cnt[WRITE][y][x]) & 0xff) * col_mul;
                    if ((blue > 255) || (red > 255)) {
                        if (blue < red) {
                            blue = blue * 255 / red;
                            red = 255;
                        } else {
                            red = red * 255 / blue;
                            blue = 255;
                        }
                    }
                    int green = (blue + red > 255) ? 0 : 255 - blue - red;
                    int tmp = (blue > red) ? 255 - blue + red : 255;
                    blue = (blue < red) ? 255 + blue - red : 255;
                    red = tmp;
                    if (cnt[DMA][y][x] != 0)
                        color[x][y] = ((red / 2) << 16) | 
                            ((128 + green / 2) << 8) | (blue / 2);
                    else
                        color[x][y] = (red << 16) | (green << 8) | blue;
                }
            }
        }
        cnt = null;
        available = true;
    }
    public int getColor(int x, int y) {
        if (color != null)
            return color[x][y];
        else
            return 0;
    }
}

/**********************************************************************/
